"""
Jacob Thompson

Models.py
"""
from copy import copy

from JT.Utils import Timer

# Downloaded Modules
import numpy as np
import matplotlib.pyplot as plot


class Model:

	def __init__(self, x, y, val_x = None, val_y = None, input_names = None, echo = True, norm = False, optimizer = None):
		self.x = x
		self.y = y
		self.val_x = val_x
		self.val_y = val_y
		self.input_names = input_names
		self.echo = echo
		self.residuals = None
		self.optimizer = optimizer
		self.timer = Timer()
		self.threshold = 0.5
	
	def Train(self, echo = None):
		if echo is None:
			echo = self.echo
		
		self.timer.tic()
		self.weights = self.optimizer.Train((self.x, self.y), echo = echo)
		self.loss = self.optimizer.loss[-1]
		
		self.elapsed = self.timer.toc()
		self.Upkeep()
		return self.weights
		
	def TrainAndValidate(self, ratio = (7, 1.5, 1.5), echo = None):
		if echo is None:
			echo = self.echo
			
		if type(echo) != type(None):
			self.echo = echo
			
		ratio = np.array(ratio)
		ratio = ratio / sum(ratio)
		indices = (ratio * self.y.shape[0]).astype(int)
		
		train_x = self.x[:indices[0]]
		train_y = self.y[:indices[0]]

		validate_x = self.x[-indices[-1]:]
		validate_y = self.y[-indices[-1]:]
		self.timer.tic()
		self.weights = self.optimizer.Train((train_x, train_y), (validate_x, validate_y), echo)
		self.elapsed = self.timer.toc()
		
		self.loss = self.optimizer.loss[-1]
		self.Upkeep()
		return self.weights
		
	def CrossValidate(self, fold_number = 5, echo = None):
		if echo is None:
			echo = self.echo
			
		if type(echo) != type(None):
			self.echo = echo
		sort_arg = np.argsort(self.y[:, 0])
		y = self.y[sort_arg, :]
		x = self.x[sort_arg, :]
	
		folds_x = []
		folds_y = []
		for i in range(fold_number):
			sample_indices = np.arange(i, x.shape[0], fold_number)
			folds_x.append(x[sample_indices, :])
			folds_y.append(y[sample_indices, :])
			
		loss_agg = 0
		acc_agg = 0
		self.Echo('Preforming Cross-Fold Validation with', fold_number, 'folds')
		self.timer.tic()
		for i in range(fold_number):
			self.Echo(f'\tFold {i+1}/{fold_number}') 
			train_x = copy(folds_x)
			train_y = copy(folds_y)
			
			val_x = train_x.pop(i)
			val_y = train_y.pop(i)
			
			train_x = np.vstack(train_x)
			train_y = np.vstack(train_y)
			
			self.weights = self.optimizer.Train((train_x, train_y), (val_x, val_y), False)
			
			self.Upkeep()
			
			loss_agg += self.optimizer.val_loss[-1]
			acc_agg += self.Accuracy()
			
			self.Echo('\t\tAverage Loss:', loss_agg / (i+1))
			self.Echo('\t\tAverage Accuracy:', acc_agg / (i+1))
		self.elapsed = self.timer.toc()
		
		self.avg_loss = loss_agg / fold_number
		self.avg_acc = acc_agg / fold_number
		return self.avg_loss
		
		
		
	def BatchTrain(self, batch_size, echo = None):
		if echo is None:
			echo = self.echo
			
		batch_indices = (np.random.rand(batch_size) * x.shape[0]).astype(int)
		batch_x = x[batch_indices]
		batch_y = y[batch_indices]
		
		self.BuildTransforms()
		batch_x = self.ApplyTransforms(batch_x)
		self.weights = self.optimizer.Train((batch_x, batch_y), echo = echo)
		self.Upkeep()
		return self.weights
		"""
		=======		WIP		=======
		"""
	
	def RSquared(self):
		if self.residuals is None:
			self.residuals = self.y - self.y_hat
		if not 'y_bar' in dir(self):	
			self.y_bar = np.mean(self.y)
		top = (self.residuals.T @ self.residuals).trace()
		bottom = ((self.y - self.y_bar).T @(self.y - self.y_bar)).trace()
		self.r_squared = 1 - top/bottom
		return self.r_squared
		
	def SSR(self):
		if self.residuals is None:
			self.residuals = self.y - self.y_hat
		self.ssr = (self.residuals.T @ self.residuals).trace()
		return self.ssr
		
	def Fit(self):
		self.y_hat = self.x @ self.weights
		return self.y_hat
		
	def Predict(self, new_data):
		if not np.ones([self.x.shape[0], 1]) in new_data:
			new_data = np.hstack([np.ones([new_data.shape, 1]), new_data])
		if new_data.shape != self.x.shape:
			raise NameError("Data Shape doesn't match model")
		prediction = new_data @ self.weights
		return prediction
		
	def Test(self, test_data, test_results):
		prediction = self.Predict(test_data)
		loss = self.loss_func.eval(self.w_hat, new_data, test_results, prediction)
		print('Loss:', loss)
		return prediction
		
	def Upkeep(self):
		self.Fit()
		self.RSquared()
		self.SSR()
		
	def Echo(self, *args):
		if self.echo:
			args = list(args)
			for i in range(len(args)):
				if type(args[i]) != 'str':
					args[i] = str(args[i])
			print(' '.join(args))
			
	def Summary(self):
		print('Optimizer:\t\t', self.optimizer.name)
		if 'loss_func' in dir(self.optimizer):
			print('Loss Function:\t', self.optimizer.loss_func.name)
			print('Iterations:\t', len(self.optimizer.loss) - 1)
		if type(self.input_names) != type(None):
			print('\t'.join(self.input_names.flatten()[:5]))
		else:
			print('\t'.join([ 'w%g' % i for i in list(np.arange(self.x.shape[1])) ][:5]))
		print('\t'.join([ '%.2f' % weight for weight in list(self.weights.flatten())][:5]))
		print('--')
		print('R^2:\t\t\t\t', self.r_squared)
		print('Loss:\t\t\t\t', self.optimizer.loss[-1])
		print('Time Elapsed:\t', round(self.elapsed, 2))
		print()
			
	### Plotting
		
	def PlotLoss(self):
		if 'loss' in dir(self.optimizer):
			plot.plot(self.optimizer.loss[2:], label = 'Model Loss')
		if 'val_loss' in dir(self.optimizer):
			plot.plot(self.optimizer.val_loss[2:], 'r', label = 'Validation Loss')
			
		plot.title('Loss Curve')
		plot.xlabel('Iterations')
		plot.ylabel('Loss')
		plot.legend()
		plot.show()
		
	def WeightHistogram(self, bins = 10, non_zero_threshold = 0):
		plot.hist(self.weights[abs(self.weights) >= non_zero_threshold], bins = bins)
		plot.show()
		
	
	def ResidualHistogram(self, bins = 10):
		if self.residuals is None:
			self.residuals = self.y - self.y_hat
		plot.hist(self.residuals, bins = bins)
		plot.show()	

	def ResidualPlot(self):
		plot.plot(self.y, self.y)
		plot.plot(self.y, self.y_hat)
		
	def ROC(self, step_size = 0.01, sig_figs = 3):
		p_hat = self.optimizer.loss_func.activation(self.y_hat)
		y = np.round(self.y) == 1
		tpr = [1]
		fpr = [1]
		min_dist = np.inf
		
		max_accuracy = 0
		max_precision = 0
		max_recall = 0
		max_f1 = 0

		auc = 0
		for t in np.arange(0, 1 + step_size, step_size):
			y_cat = p_hat >= t
			tp = np.sum(y & y_cat)
			fp = np.sum((y == False) & y_cat)
			tn = np.sum((y == False) & (y_cat == False))
			fn = np.sum(y & (y_cat == 0))
			
			tpr.append(tp / (tp + fn))
			fpr.append(fp / (fp + tn))
			
			dist = (np.sqrt((1 - tpr[-1]) ** 2 + fpr[-1] ** 2))
			accuracy = (tp + tn) / (tp + tn + fp + fn)
			precision = tp / (tp + fp)
			recall = tp / (tp + fn)
			f1 = 2 * precision * recall / (precision + recall)
			
			
			if dist < min_dist:
				min_dist = dist
				opt_t = t
			# if accuracy > max_accuracy:
				# max_accuracy = accuracy
				# opt_t_accuracy = t
			if f1 > max_f1:
				max_f1 = f1
				opt_t_f1 = t
			

			auc += tpr[-1] * (fpr[-2] - fpr[-1])
		
		opt_i = int(opt_t / step_size)
		# opt_i_accuracy = int(opt_t_accuracy / step_size)
		opt_i_f1 = int(opt_t_f1 / step_size)
		plot.plot(fpr, tpr)
		plot.plot([0, 1], [0, 1], 'k:')
		plot.scatter(fpr[opt_i], tpr[opt_i], c='g', label = 'Closest')
		# plot.scatter(fpr[opt_i_accuracy], tpr[opt_i_accuracy], c='r')
		plot.scatter(fpr[opt_i_f1], tpr[opt_i_f1], c='m', label = 'F1 Optimal')
		plot.title('ROC Curve')
		plot.xlabel('False Positive Rate')
		plot.ylabel('True Positive Rate')
		plot.legend()
		print('Area Under Curve:\t', round(auc, sig_figs))
		print('Optimal Threshold:')
		print('\tOverall:', opt_t)
		# print('\tAccuracy:', opt_t_accuracy)
		print('\tF1 Score:', opt_t_f1)
		print('At overall threshold:')
		print('\tAccuracy:', round(self.Accuracy(opt_t), sig_figs))
		print('\tPrecision:', round(self.Precision(opt_t), sig_figs))
		print('\tRecall:', round(self.Recall(opt_t), sig_figs))
		print('\tF1 Score:', round(self.F1Score(opt_t), sig_figs))
		print('At F1 threshold:')
		print('\tAccuracy:', round(self.Accuracy(opt_t_f1), sig_figs))
		print('\tPrecision:', round(self.Precision(opt_t_f1), sig_figs))
		print('\tRecall:', round(self.Recall(opt_t_f1), sig_figs))
		print('\tF1 Score:', round(self.F1Score(opt_t_f1), sig_figs))
		self.threshold = opt_t_f1
		return auc
	
	def OutcomeMatrix(self, threshold = None):
		if threshold is None:
			threshold = self.threshold
			
		p_hat = self.optimizer.loss_func.activation(self.y_hat)
		y_cat = p_hat >= threshold
		y = np.round(self.y) == 1
		
		tp = np.sum(y & y_cat)
		fp = np.sum((y == False) & y_cat)
		tn = np.sum((y == False) & (y_cat == False))
		fn = np.sum(y & (y_cat == 0))
		return tp, tn, fp, fn
	
	def Accuracy(self, threshold = None):
		if threshold is None:
			threshold = self.threshold
		tp, tn, fp, fn = self.OutcomeMatrix(threshold)
		return (tp + tn) / (tp + tn + fp + fn)
		
	def Precision(self, threshold = None):
		if threshold is None:
			threshold = self.threshold
		tp, tn, fp, fn = self.OutcomeMatrix(threshold)
		return tp / (tp + fp)
	
	def Recall(self, threshold = None):
		if threshold is None:
			threshold = self.threshold
		tp, tn, fp, fn = self.OutcomeMatrix(threshold)
		return tp / (tp + fn)
		
	def F1Score(self, threshold = None):
		if threshold is None:
			threshold = self.threshold
		tp, tn, fp, fn = self.OutcomeMatrix(threshold)
		precision = tp / (tp + fp)
		recall = tp / (tp + fn)
		return 2 * precision * recall / (precision + recall)
		
	def GetHyperParameters(self):
		return self.optimizer.GetHyperParameters()
		
	def SetHyperParameters(self, param_dict):
		self.optimizer.SetHyperParameters(param_dict)
		
	def HyperTune(self, param_dict, fold_number = 5, metric = 'loss'):
		exec_dict = {}
		self.best_metric = np.inf
		def RecTune(param_list):
			param_pair = param_list[0]
			for param in param_pair[1]:
				exec_dict[param_pair[0]] = param
				if len(param_list) == 1:
					print(exec_dict)
					self.SetHyperParameters(exec_dict)
					try:
						self.CrossValidate(fold_number)
					except MemoryError:
						self.TrainAndValidate()
					print(getattr(self, metric))
					if self.loss < self.best_metric:
						self.best_metric = copy(self.loss)
						self.best_dict = copy(exec_dict)
				else:
					RecTune(param_list[1:])
			
		param_list = []
		for key, value in param_dict.items():
			param_list.append([key, value])
		
		RecTune(param_list)
		return self.best_dict
		
			
		
		
		
	


'''
Jacob Thompson

Optimizers.py
'''
import numpy as np
import JT.LossFunctions


class Optimizer:
	def __init__(self):
		self.hyperparameters = {}
	
	def Echo(self, *args):
		if self.echo:
			args = list(args)
			for i in range(len(args)):
				if type(args[i]) != 'str':
					args[i] = str(args[i])
			print(' '.join(args))
	
	def SetHyperParameters(self, param_dict):
		next_dict = dict()
		for key, value in param_dict.items():
			if key in self.hyperparameters:
				setattr(self, key, value)
			else:
				next_dict[key] = value
		self.loss_func.SetHyperParameters(next_dict)
		
	def GetHyperParameters(self):
		out = {}
		for param in self.hyperparameters:
			out[param] = getattr(self, param)
		out.update(self.loss_func.GetHyperParameters())
		return out
			

class Linear_Regression(Optimizer):
	def __init__(self, lamb = 0, echo = False):
		self.name = 'Linear Regression'
		self.lamb = lamb
		self.loss_func = JT.LossFunctions.Loss()
		self.hyperparameters = {'lamb'}
		
	
	def Train(self, x, y, echo = None):
		self.weights = np.linalg.inv(x.T @ x + np.eye(x.shape[1]) * self.lamb) @ x.T @ y	# Calculates weights based on the closed form formula for linear regression
		return self.weights
		
	

class Gradient_Descent(Optimizer):
	"""
		Runs Gradient Descent on the data
		
			learn_rate: 				learning rate of the optimizer (eta)
			stop_difference: 	The minimum difference between each epoch's loss that will keep the optimizer running
			max_epochs: 		The maximum allowed epochs (the optimizer will stop after this many iterations)
			loss_func:			The loss function used for optimization. Loss functions can be found in LossFunctions.py
			
		Returns the calculated weights
	"""
	
	def __init__(self, learn_rate, stop_difference = 1e-3, max_epochs = 1e3, loss_func = JT.LossFunctions.SumOfSquaredResidual(), echo = False):
		self.name = 'Gradient Descent'
		self.learn_rate = learn_rate
		self.stop_difference = stop_difference
		self.max_epochs = max_epochs
		self.loss_func = loss_func
		self.echo = echo
		self.hyperparameters = {'learn_rate', 'stop_difference', 'max_epochs'}
	
	def Train(self, training_set, validation_set = None,  echo = None, plot_loss = False, init = True):
		'''
			training_set: (x, y)
			validation_set: [validation_x, validation_y]
			
			set init to False if you want the weights/loss to be kept every time this optimizer is run
		'''
		if type(echo) != type(None):
			self.echo = echo
			
		x, y = training_set
		
		validate = False
		if type(validation_set) != type(None):
			val_x, val_y = validation_set
			if type(val_x) != type(None) and type(val_y) != type(None):
				validate = True
				if not 'val_loss' in dir(self) or init:
					self.val_loss = [np.inf]
		
		if not 'weights' in dir(self) or init:																	# Checks if the optimizer already has weights
			self.weights = np.random.randn(x.shape[1], y.shape[1])										# If it doesn't, initializes random weights
		if not 'loss' in dir(self) or init:
			self.loss = [0, np.inf]
		
		i = 1
		check = np.inf
		while (check > self.stop_difference) and (i < self.max_epochs):	# While loop that stops when either the max epochs have elapsed or the stop difference is met
			y_hat = x @ self.weights																	# Calculates y_hat from the current weights
			gradient = self.loss_func.Derivative(self.weights, x, y, y_hat)								# Calcuates the gradient by calling the derivative of the loss function
			self.weights -= self.learn_rate * gradient														# Adjusts the weights based on the gradient and the training rate
			self.loss.append(self.loss_func.Evaluate(y, y_hat, self.weights))						# Calculates loss and keeps track of it
			if validate:
				val_y_hat = val_x @ self.weights
				self.val_loss.append(self.loss_func.Evaluate(val_y, val_y_hat, self.weights))
				check = abs(self.val_loss[i] - self.val_loss[i-1])
			else:
				check = abs(self.loss[i] - self.loss[i-1])
			
			self.Echo('Iteration:', i, '\tLoss:', self.loss[-1])										# If self.echo is True, will print the latest loss
			i += 1
			
		return self.weights
		
	
	
class Batch_Gradient_Descent(Optimizer):
	def __init__(self, learn_rate, batch_size = 10, stop_difference = 1e-5, max_epochs = 1e3, loss_func = JT.LossFunctions.SumOfSquaredResidual(), echo = False):
		self.name = 'Batch Gradient Descent'
		self.learn_rate = learn_rate
		self.batch_size = batch_size
		self.stop_difference = stop_difference
		self.max_epochs = max_epochs
		self.loss_func = loss_func
		self.echo = echo
		self.hyperparameters = {'learn_rate', 'stop_difference', 'max_epochs', 'batch_size'}
	
	def Train(self, training_set, validation_set = None,  echo = None):
		'''
		Runs Gradient Descent on the data
		
			learn_rate: 				learning rate of the optimizer (eta)
			stop_difference: 	The minimum difference between each epoch's loss that will keep the optimizer running
			max_epochs: 		The maximum allowed epochs (the optimizer will stop after this many iterations)(
			loss_func:			The loss function used for optimization. Loss functions can be found in LossFunctions.py
			
		Calls self.Upkeep to update everything
		Returns the calculated weights
		'''
		
		if type(echo) != type(None):
			self.echo = echo
			
		x, y = training_set
		if not 'loss' in dir(self):
			self.loss = [0, np.inf]
		
		validate = False
		if type(validation_set) != type(None):
			val_x, val_y = validation_set
			if type(val_x) != type(None) and type(val_y) != type(None):
				validate = True
				if not 'val_loss' in dir(self):
					self.val_loss = [np.inf]
		
		if not 'weights' in dir(self):																	# Checks if the optimizer already has weights
			self.weights = np.random.randn(x.shape[1], y.shape[1])												# If it doesn't, initializes random weights
		
		i = 1
		check = np.inf
		while (check > self.stop_difference) and (i < self.max_epochs):	# While loop that stops when either the max epochs have elapsed or the stop difference is met
			batch_indices = (np.random.rand(self.batch_size) * x.shape[0]).astype(int)
			batch_x = x[batch_indices]
			batch_y = y[batch_indices]
			batch_y_hat = batch_x @ self.weights															# Calculates y_hat from the current weights
			gradient = self.loss_func.Derivative(self.weights, batch_x, batch_y, batch_y_hat)				# Calcuates the gradient by calling the derivative of the loss function
			self.weights -= self.learn_rate * gradient													# Adjusts the weights based on the gradient and the training rate
			self.loss.append(self.loss_func.Evaluate(batch_y, batch_y_hat, self.weights))		# Calculates loss and keeps track of it
			
			if validate:
				val_y_hat = val_x @ self.weights
				self.val_loss.append(self.loss_func.Evaluate(val_y, val_y_hat, self.weights))
				check = abs(self.val_loss[i] - self.val_loss[i-1])
			else:
				check = abs(self.loss[i] - self.loss[i-1])
			self.Echo('Iteration:', i, '\tLoss:', self.loss[-1])										# If self.echo is True, will print the latest loss
			i += 1
		return self.weights

class Stochastic_Gradient_Descent(Batch_Gradient_Descent):
	def __init__(self, learn_rate, stop_difference = 1e-5, max_epochs = 1e6, loss_func = JT.LossFunctions.SumOfSquaredResidual(), echo = False):
		self.name = 'Stochastic Gradient Descent'
		self.learn_rate = learn_rate
		self.batch_size = 1
		self.stop_difference = stop_difference
		self.max_epochs = max_epochs
		self.loss_func = loss_func
		self.echo = echo
		self.hyperparameters = {'learn_rate', 'stop_difference', 'max_epochs'}




import torch
import copy

from JT.Utils import RecGetattr, RecSetattr

def Breed(parent_1, parent_2, p_1, p_mutate):
    """
    Breeds two tensors.

    Elements are randomly selected from each parent tensor to be passed on to the child tensor.

    Some of the elements are set to zero to "mutate" the child a bit.

    Note: Tensors must be the same shape.

    :param parent_1: The first parent tensor.
    :param parent_2: The second parent tensor.
    :param p_1: Proportion of the child that should come from the first parent.
    :param p_mutate: Proportion of the child to be set to zero (mutated).
    :return: The child tensor.
    """
    rng = torch.rand(parent_1.shape)
    bool_1 = rng < p_1
    bool_2 = rng > (p_1 + p_mutate)
    return parent_1 * bool_1.float() + parent_2 * bool_2.float()


def ModelBreed(model_1, model_2, mutation_rate):
    """
    Breeds two models.
    The models must have been built using the torch.nn.Module framework.

    Each parameter in the models are bred together using Breed().

    The influence of each parent is derived from fitness and performance attributes in the models.

    Note: The models need to be exactly the same type.

    :param model_1: The first parent model.
    :param model_2: The second parent model.
    :param mutation_rate: The mutation rate.
    :return: The child model.
    """
    model_child = copy.copy(model_1)
    p_0 = (1 - model_1.performance - model_2.performance) * mutation_rate
    p_male = model_1.fitness / (model_1.fitness + model_2.fitness) * (1 - p_0)
    for name, param in model_child.named_parameters():
        child = torch.nn.Parameter(Breed(RecGetattr(model_1, name), RecGetattr(model_2, name), p_male, p_0))
        RecSetattr(model_child, name, child)
    return model_child


import torch
from torch.nn import CrossEntropyLoss


DTYPE = torch.FloatTensor

class Layer:
    def __init__(self, size_in, size_out, activation):
        self.weights = torch.randn(size_in, size_out, type=DTYPE, requires_grad=True)
        self.bias = torch.randn(1, size_out, type=DTYPE, requires_grad=True)
        self.activation = activation

    def __call__(self, z_in):
        return self.activation(self.weights @ z_in + self.bias)

    def GetVariables(self):
        return self.weights, self.bias


class LSTM:
    def __init__(self, rate, input_size, short_size, long_size, output_size):
        self.input_size = input_size
        self.short_size = short_size
        self.long_size = long_size
        self.inner_size = self.input_size + self.short_size
        self.output_size = output_size

        self.forget_gate = Layer(self.inner_size, self.long_size, torch.sigmoid)

        self.memory_layer = Layer(self.inner_size, self.long_size, lambda x: x)
        self.memory_gate = Layer(self.inner_size, self.long_size, torch.sigmoid)

        self.recall_layer = Layer(self.long_size, self.short_size, torch.tanh)
        self.recall_gate = Layer(self.inner_size, self.short_size, torch.sigmoid)

        self.output_layer = Layer(self.short_size, self.output_size, lambda x: x)

        self.optimizer = torch.optim.Adam(self.GetVariables(), lr=rate)
        self.loss_func = torch.nn.CrossEntropyLoss()

    def GetVariables(self):
        self.variables = []
        for name in dir(self):
            obj = getattr(self, name)
            if type(obj) == Layer:
                self.variables.extend(obj.GetVariables())
        return self.variables

    def Forward(self, x, constant):
        mem_short = torch.zeros(1, self.short_size)
        mem_long = torch.zeros(1, self.long_size)

        out_list = []

        for i in range(x.shape[0] - 1):
            inner = torch.cat((x[[i], :], mem_short, constant), dim=1)
            mem_long = mem_long * self.forget_gate(inner)
            mem_long = mem_long + self.memory_gate(inner) * self.memory_layer(inner)
            mem_short = self.recall_gate(inner) * self.recall_layer(mem_long)
            out = self.output_layer(mem_short)
            out_list.append(out)

        return torch.cat(out_list)

    def Backward(self, y_hat, y):
        loss = self.loss_func(y_hat, y)
        loss.backward()
        self.optimizer.step()
        self.optimizer.zero_grad()
        return loss.detach()

    def Train(self, x, constant, y):
        y_hat = self.Forward(x, constant)
        loss = self.Backward(y_hat, y)
        return y_hat.detach().cpu().numpy(), loss.detach().cpu().numpy()

    def Generate(self, constant, start_token, stop_token, factor=0.001, ):
        mem_short = torch.randn(1, self.short_size) * factor
        mem_long = torch.randn(1, self.long_size) * factor

        out = ''
        out_list = []
        while out != stop_token:
            inner = torch.cat((start_token, mem_short, constant), dim=1)
            mem_long = mem_long * self.forget_gate(inner)
            mem_long = mem_long + self.memory_gate(inner) * self.memory_layer(inner)
            mem_short = self.recall_gate(inner) * self.recall_layer(mem_long)
            out = self.output_layer(mem_short)
            out_list.append(out)
        return out_list








import numpy as np

def Eigen(x):
    eigen_values, eigen_vectors = np.linalg.eig(x.T @ x)
    sort_indices = np.argsort(eigen_values)[::-1]
    return eigen_values[sort_indices], eigen_vectors[:, sort_indices]

class PCA:
    def Train(self, x):
        self.eigen_values, self.eigen_vectors = np.linalg.eig(x.T @ x)
        self.sort_indices = np.argsort(self.eigen_values)[::-1]

    def Encode(self, x, components):
        return x @ self.eigen_vectors[:, self.sort_indices[:components]]

    def Decode(self, x):
        return x @ self.eigen_vectors[:, self.sort_indices[:x.shape[-1]]].T

    def Filter(self, x, components):
        return self.Decode(self.Encode(x, components))






name = 'JT'

import JT.Models
import JT.LossFunctions
import JT.Optimizers
import JT.Transforms
import JT.MakeSet
import JT.NNet
import JT.Layers
import JT.Initializers
import JT.Utils
import JT.Gradients
import JT.Clustering
import JT.Trees


import numpy as np
import scipy as sp
from scipy import stats

# TODO: Write this entire module

'''
## Standard K-NN
y_hat = []
k = 0
for point in x_sim:
    diff = x - point
    dist = np.sqrt(np.sum(diff * diff, axis = 1))
    nearest = np.argsort(dist)[:k]
    cat = sp.stats.mode(np.argmax(y[nearest], axis = 1))
    y_hat.append(cat.mode[0])
plot.scatter(x_sim[:, 0], x_sim[:, 1], c = y_hat)
'''

def Nothing(dist):
    return dist

class KNearest:
    def __init__(self, x, y, method = None):
        self.x = x
        self.y = y
        self.method = method

    def Predict(self, x, k):
        y_hat = []
        for point in x:
            diff = self.x - point
            dist = np.sqrt(np.sum(diff * diff, axis = 1))
            nearest = np.argsort(dist)[:k]
            w = self.method(dist[nearest])


"""
Jacob Thompson

Clustering.py
"""

import numpy as np
import scipy as sp
import scipy.stats
from matplotlib import pyplot as plot

from collections import defaultdict
import warnings

from JT.Utils import ProgressBar

def FindElbow(y):
    slope = np.diff(y)
    slope_0 = np.hstack([slope, slope[-1]])
    slope_1 = np.hstack([slope[0], slope])
    out = np.arctan(-abs(1 / slope_0)) + np.arctan(-abs(slope_1)) + np.pi / 2
    out[np.where(np.isnan(out))] = 1
    return np.argmin(out)

def GetCentroids(x, cats, centroids):
    for i in range(centroids.shape[0]):
        if x[np.where(cats == i)].size:
            centroids[i] = np.mean(x[np.where(cats == i)], axis = 0)
    return centroids

# Silhouette Analysis
def SilhouetteValues(x, cats, centroids):
    silhouette_values = np.ones(cats.shape) * -1
    for i in range(centroids.shape[0]):
        b = np.sum((x - centroids[i])[np.where(cats == i)] ** 2) / np.sum(cats == i)
        a = 0
        for j in range(centroids.shape[0]):
            if j != i:
                a += np.sum((x - centroids[j])[np.where(cats == i)] ** 2, axis = 1)
        if centroids.shape[0] > 1:
            a /= centroids.shape[0] - 1
            silhouette_values[np.where(cats == i)] = (a - b) / np.max(np.vstack([a, np.ones(a.shape) * b]), axis=0)
    return silhouette_values

def SilhouettePlot(x, cats, centroids):
    silhouette_values = SilhouetteValues(x, cats, centroids)
    y = []
    for i in range(centroids.shape[0]):
        y_i = silhouette_values[np.where(cats == i)]
        y_i.sort()
        y_i = np.flip(y_i)
        y.append(y_i)
    y = np.hstack(y)
    cats.sort()
    plot.scatter(np.arange(y.shape[0]), y, c=cats)

# Loss Functions
class HardLoss:
    def __init__(self):
        self.name = 'Average Distance Loss'
        self.optimization_method = FindElbow

    def __call__(self, x, cats, centroids):
        cost = 0
        for i in range(len(centroids)):
            diff = (x - centroids[i])[np.where(cats == i)]
            dist = (np.sum(diff * diff, axis=1, keepdims=True))
            cost += np.sum(dist)
        return cost

class SoftLoss:
    def __init__(self, beta):
        self.name = 'Responsibility Distance Loss'
        self.beta = beta
        self.optimization_method = FindElbow

    def __call__(self, x, cats, centroids):
        cost = 0
        for i in range(centroids.shape[0]):
            if np.sum(np.where(cats == i)) > 0:
                # print('sum:', np.sum(np.where(cats == i)))
                # print('centroids[i]:', centroids[i])
                diff = (x - centroids[i])[np.where(cats == i)]
                # print('diff:', diff)
                dist = (np.sum(diff * diff, axis=1, keepdims=True))
                # print('dist:', dist)
                a = np.exp(-self.beta * dist)
                # print('a:', np.sum(a / (sum(a) + 1e-40)))
                responsibility = a / (np.sum(a) + 1e-40)
                # print('responsibility:', responsibility)
                cost += np.sum(responsibility * dist)
                # print(cost)

        return cost / centroids.shape[0]

class DBI:
    def __init__(self):
        self.optimization_method = np.argmin
        self.name = 'Davies Bouldin Index'

    def __call__(self, x, cats, centroids):
        def Sig(i):
            n = np.sum(cats == i)
            return np.sum((x - centroids[i])[np.where(cats == i)] ** 2) / n

        out = 0
        for k in range(centroids.shape[0]):
            inside_max = 0
            sig_k = Sig(k)
            for j in range(centroids.shape[0]):
                if j != k:
                    sig_j = Sig(j)
                    dist = np.sum((centroids[k] - centroids[j]) ** 2)
                    inside = (sig_k + sig_j) / dist
                    if inside > inside_max:
                        inside_max = inside

            out += inside_max
        if out == 0:
            return np.inf
        else:
            return out / centroids.shape[0]
class CHI:
    def __init__(self):
        self.optimization_method = np.argmax
        self.name = 'Calinski-Haribasz Index'

    def __call__(self, x, cats, centroids):
        total_mean = np.mean(x, axis=0, keepdims=True)
        top = 0
        bot = 0
        for k in range(centroids.shape[0]):
            n = np.sum(cats == k)
            if n > 0:
                top += np.sum((centroids[k] - total_mean) ** 2) * n
                bot += np.sum((x - centroids[k])[np.where(cats == k)] ** 2)
        return (top / bot) * (x.shape[0] - centroids.shape[0]) / (centroids.shape[0])

class GaussianLoss:
    def __init__(self):
        self.optimization_method = FindElbow
        self.name = 'Gaussian Mixture Loss'

    def __call__(self, x, cats, centroids):
        densities = []
        pis = []
        for i in range(centroids.shape[0]):
            if np.sum(cats == i) <= 1:
                density = np.zeros((x.shape[0]))
                density[np.where(cats == i)] = 1
            else:
                covariance = np.cov(x[np.where(cats == i)].T)
                gaussian = (sp.stats.multivariate_normal(mean=centroids[i], cov=covariance, allow_singular=True))
                density = gaussian.pdf(x)

            pi = np.sum(cats == i) / x.shape[0]
            pis.append(pi)
            densities.append(density)

        densities = np.stack(densities).T
        pis = np.stack(pis)
        return -np.sum(np.log(np.sum(pis * densities, axis = 1) + 1e-99))

class SilhouetteLoss:
    def __init__(self):
        self.optimization_method = FindElbow
        self.name = 'Average Silhouette Value'

    def __call__(self, x, cats, centroids):
        return np.mean(SilhouetteValues(x, cats, centroids))


class SilhouetteLogLoss:
    def __init__(self):
        self.optimization_method = FindElbow
        self.name = 'Average Log Silhouette Value'

    def __call__(self, x, cats, centroids):
        return np.mean(np.log(SilhouetteValues(x, cats, centroids) + 1.0001))

# Clustering Methods
def HardKMeans(x, num=3):
    warnings.warn("deprecated: use the class method Hard_K_Means instead", DeprecationWarning)
    centroids = x[np.random.choice(np.arange(x.shape[0]), num), :]
    cent_diff = np.inf
    iteration = 0
    while cent_diff > 0.001 and iteration < 11:
        iteration += 1
        cent_diff = 0
        dists = []
        for centroid in centroids:
            diff = x - centroid
            dist = np.sqrt(np.sum(diff * diff, axis=1, keepdims=True))
            dists.append(dist)

        dists = np.hstack(dists)
        cats = np.argmin(dists, axis=1)
        cost = 0
        for i in range(centroids.shape[0]):
            centroid = np.mean(x[np.where(cats == i)], axis=0)
            cent_diff += np.sum(np.abs(centroid - centroids[i]))
            centroids[i] = centroid

            diff = (x - centroid)[np.where(cats == i)]
            dist = np.sqrt(np.sum(diff * diff, axis=1, keepdims=True))
            cost += np.sum(dist)

    return np.array([cats]).T, centroids, cost

class Hard_K_Means:
    def __init__(self):
        self.loss_func = HardLoss()
    def GetDistances(self, x):
        distances = []
        for centroid in self.centroids:
            difference = x - centroid
            distance = np.sqrt(np.sum(difference * difference, axis=1, keepdims=True))
            distances.append(distance)
        return np.hstack(distances)

    def Predict(self, x):
        return np.argmax(self.GetDistances(x), axis = 1)

    def Train(self, x, clusters, max_iterations = 1000, stop_difference = 1e-6):
        self.centroids = x[np.random.choice(np.arange(x.shape[0]), clusters), :]
        self.losses = [np.inf]

        for iteration in range(max_iterations):
            y_hat = self.Predict(x)
            self.centroids = GetCentroids(x, y_hat, self.centroids)
            self.losses.append(self.loss_func(x, y_hat, self.centroids))
            loss_diff = abs((self.losses[-1] - self.losses[-2]) / self.losses[-1])
            print(self.losses[-1])
            print('\rHard K-Means | Iterations: {:>7} | Loss Difference: {:>9,.4} | Loss: {:>9,.4}'.format(iteration, loss_diff, self.losses[-1]),ProgressBar(np.log(loss_diff) / np.log(stop_difference)), end = '')
            if loss_diff < stop_difference:
                break

        print()
        y_hat = self.Predict(x)
        return y_hat, self.centroids, self.losses[-1]


def SoftKMeans(x, beta=1., num=3, loss_func=HardLoss(), loss_threshold = 1e-6):
    warnings.warn("deprecated: use the class method Soft_K_Means instead", DeprecationWarning, stacklevel = 2)
    centroids = x[np.random.choice(np.arange(x.shape[0]), num), :]

    loss_new = 1.
    loss_diff = np.inf
    iteration = 0
    # print('\rSoft K-Means | Iterations: {:>7} | Loss Difference: {:>9,.4} | Loss: {:>9,.4}'.format(iteration, loss_diff, loss_new), ProgressBar(0 / np.log(loss_threshold)), end = '')
    while loss_diff > loss_threshold and iteration < 100:

        loss_old = loss_new
        iteration += 1
        responsibilities = []
        for centroid in centroids:
            diff = x - centroid
            dist = (np.sum(diff * diff, axis=1, keepdims=True))
            responsibility = np.exp(-beta * dist) / np.sum(np.exp(-beta * dist))
            responsibilities.append(responsibility)
        responsibilities = np.hstack(responsibilities)

        cats = np.argmax(responsibilities, axis=1, )
        centroids = GetCentroids(x, cats, centroids)

        loss_new = loss_func(x, cats, centroids)
        loss_diff = abs((loss_new - loss_old) / loss_new)
        print('\rSoft K-Means | Iterations: {:>7} | Loss Difference: {:>9,.4} | Loss: {:>9,.4}'.format(iteration, loss_diff, float(loss_new)), ProgressBar(np.log(loss_diff) / np.log(loss_threshold)), end = '')
    print()
    return cats, centroids, loss_new

class Soft_K_Means:
    def __init__(self, beta = 1.):
        self.beta = beta
        self.loss_func = HardLoss()#beta = self.beta)

    # def GetCentroids(self, x):

    def GetResponsibilities(self, x):
        responsibilities = []
        for centroid in self.centroids:
            diff = x - centroid
            dist = (np.sum(diff * diff, axis=1, keepdims=True))
            responsibility = np.exp(-self.beta * dist) / np.sum(np.exp(-self.beta * dist))
            responsibilities.append(responsibility)
        # print(np.hstack(responsibilities).shape)
        return np.hstack(responsibilities)

    def Predict(self, x):
        return np.argmax(self.GetResponsibilities(x), axis=1)

    def Train(self, x, clusters, max_iterations = 1000, stop_difference = 1e-6):
        self.centroids = x[np.random.choice(np.arange(x.shape[0]), clusters), :]
        self.losses = [np.inf]
        print('\rSoft K-Means | Iterations: {:>7} | Loss Difference: {:>9,.4} | Loss: {:>9,.4}'.format(0, np.inf, self.losses[-1]), ProgressBar(0 / np.log(stop_difference)), end = '')
        for iteration in range(max_iterations):
            y_hat = self.Predict(x)
            self.centroids = GetCentroids(x, y_hat, self.centroids)
            print(x.shape)
            self.losses.append(self.loss_func(x, y_hat, self.centroids))
            loss_diff = abs((self.losses[-1] - self.losses[-2]) / self.losses[-1])
            # print(self.losses[-1])
            print('\rSoft K-Means | Iterations: {:>7} | Loss Difference: {:>9,.4} | Loss: {:>9,.4}'.format(iteration, loss_diff, float(self.losses[-1])), ProgressBar(np.log(loss_diff) / np.log(stop_difference)), end = '')
            if loss_diff < stop_difference:
                break
        print()
        y_hat = self.Predict(x)
        return y_hat, self.centroids, self.losses[-1]


class Gaussian_Mixture:
    def __init__(self):
        self.loss_func = GaussianLoss()

    def Train(self, x, clusters, max_iterations = 1000, stop_difference = 1e-6, prime = True):
        if prime:
            primer = Soft_K_Means()
            y_hat, self.centroids, loss = primer.Train(x, clusters = clusters, stop_difference = 1e-2)
        else:
            self.centroids = x[np.random.choice(np.arange(x.shape[0]), clusters),:]
        m = np.random.rand(clusters, x.shape[1], x.shape[1])
        self.covariances = m @ m.transpose((0, 2, 1))
        # self.covariances = np.stack([np.eye(x.shape[1])] * clusters)
        self.losses = [np.inf]
        for iteration in range(max_iterations):

            y_hat = self.Predict(x)

            self.losses.append(self.loss_func(x, y_hat, self.centroids))

            loss_diff = abs((self.losses[-1] - self.losses[-2]) / self.losses[-1])
            iteration += 1
            print('\rGaussian Mixture | Iterations: {:>7} | Loss Difference: {:>9,.4} | Loss: {:>9,.4}'.format(iteration, loss_diff, self.losses[-1]), ProgressBar(np.log(loss_diff) / np.log(stop_difference)), end = '')
            if loss_diff < stop_difference:
                break
            covariances = []
            for cluster in range(clusters):
                if x[np.where(y_hat == cluster)].size:
                    self.centroids[cluster] = np.mean(x[np.where(y_hat == cluster)], axis=0)

                covariance = np.cov(x[np.where(y_hat == cluster)].T)
                if np.isnan(covariance).any():
                    covariance = np.eye(covariance.shape[0])
                covariances.append(covariance)

            self.covariances = np.stack(covariances)

        pis = []
        for cluster in range(clusters):
            pi = np.sum(y_hat == cluster) / x.shape[0]
            pis.append(pi)
        self.pis = np.stack(pis)
        y_hat = self.Predict(x)
        print()
        return y_hat

    def Predict(self, x):
        return np.argmax(self.GetDensities(x), axis = 0)

    def GetDensities(self, x):
        self.densities = []
        for i in range(self.centroids.shape[0]):
            gaussian = (
                    sp.stats.multivariate_normal(mean=self.centroids[i], cov=self.covariances[i], allow_singular=True))
            density = gaussian.pdf(x)
            self.densities.append(density)
        return np.stack(self.densities)

    def Compress(self, x):
        y_hat = self.Predict(x)
        return self.centroids[y_hat]

    def Generate(self, num):
        sample_sizes = np.random.multinomial(num, pvals=self.pis)

        samples = []
        for i, clusters in enumerate(sample_sizes):
            samples.append(np.random.multivariate_normal(mean=self.centroids[i, :], cov=self.covariances[i, :, :], size=clusters))
        samples = np.vstack(samples)
        return samples

# Analysis

def ClusterAnalysis(x, start = 1, stop = 25, attempts = 5, beta = 0.1, loss_functions = (HardLoss(), DBI(), CHI(), SilhouetteLoss(), SilhouetteLogLoss(), GaussianLoss())):

    losses_dict = defaultdict(list)

    for i in range(start, stop):
        cats, centroids, cost = SoftKMeans(x, beta, i)
        for loss_func in loss_functions:
            nans = 0
            loss_agg = 0
            for _ in range(attempts):
                try:
                    loss = loss_func(x, cats, centroids)
                    if np.isnan(loss):
                        nans += 1
                    else:
                        loss_agg += loss
                except:
                    nans += 1
            if (attempts - nans) > 0:
                loss_avg = loss_agg / (attempts - nans)
            else:
                loss_avg = np.nan
            losses_dict[loss_func.name].append(loss_avg)

    plot.figure(figsize = (10, 5 * len(loss_functions)))

    predictions = []
    i = 0

    for loss_func in loss_functions:
        i += 1
        losses = losses_dict[loss_func.name]

        predictions.append(loss_func.optimization_method(losses))

        plot.figure(figsize = (10, 5))
        plot.plot(np.arange(1, len(losses) + 1), losses)
        plot.scatter(predictions[-1] + 1, losses[predictions[-1]], c='r', marker='X')
        plot.title(loss_func.name)
        plot.xlim(1, len(losses))

    return sp.stats.mode(predictions)



class Gaussian_Mixture_Model:
    def __init__(self, x):
        self.x = x
        self.loss_func = GaussianLoss()

    # def Train(self, clusters):













"""
Jacob Thompson

Utils
"""
import numpy as np
from matplotlib import pyplot as plot
from skimage.transform import resize
import time

class Timer:
    def __init__(self):
        self.tictoc = time.time()

    def tic(self):
        self.tictoc = time.time()

    def toc(self):
        return time.time() - self.tictoc


def ImagePad(image, padding):
    image_padded = np.pad(image, ((padding, padding), (padding, padding), (0, 0)), mode='constant',
                        constant_values=0)
    return image_padded

def LoadImage(path, size = None):
    image = plot.imread(path).copy()
    if not size is None:
        image = resize(image, output_shape = size)

    image = np.swapaxes(image, 0, 2)
    image[0] = np.flip(image, 0)
    return image

def ProgressBar(percent, size = 15):
    """
    Generates a progress bar string that is useful for tracking progress.

    :param percent: Percentage of the bar to be filled in.
    :param size: Size of the bar (how many characters long should the bar be?).
    :return: A string containing the progress bar.
    """
    done_chars = int(min(max(percent, 0), 1) * size)
    left_chars = size - done_chars
    return '|' + chr(9608) * done_chars + ' ' * left_chars + '|'


class TimeList:
    def __init__(self, list_in=None):
        """
        This is a simple wrapper class for a list that will also keep track of the time when an element was added to the
        list.

        This class has three primary attributes:
            TimeList.list: The actual values of the list.
            TimeList.time: The time that each value was added to the list.
            TimeList.rel_time: The time that each value was added to the list since the list was created.

        :param list_in: Initial iterable to populate the list
        """
        self.start_time = time.time()
        if list_in:
            self.list = list(list_in)
            self.time = [self.start_time] * len(self.list)
            self.rel_time = [0.] * len(self.list)
        else:
            self.list = []
            self.time = []
            self.rel_time = []

    def append(self, item):
        """
        Appends an item to the list and keeps track of the time that it was added.

        :param item: Item to append to the list
        :return:
        """
        current = time.time()
        self.time.append(current)
        self.rel_time.append(current - self.start_time)
        self.list.append(item)


def RecGetattr(obj, name):
    """
    A beefier version of the builtin getattr() that can get sub attributes - attributes of attributes.

    Accomplishes this by recursively calling itself.

    :param obj: Object to get the attribute from.
    :param name: Name of the attribute to get including any parent attributes. Should be in the form of: "parent.parent.attribute"
    :return: The target attribute.
    """
    name_lst = name.split('.')
    new_obj = getattr(obj, name_lst[0])
    if len(name_lst) > 1:
        return RecGetattr(new_obj, '.'.join(name_lst[1:]))
    else:
        return new_obj


def RecSetattr(obj, name, value):
    """
    A beefier version of the builtin setattr() that can set sub attributes - attributes of attributes.

    Accomplishes this by recursively calling itself.

    :param obj: Object containing the attribute to be set.
    :param name: Name of the attribute to set including any parent attributes. Should be in the form of: "parent.parent.attribute"
    :param value: Value to set the attribute to
    :return:
    """
    name_lst = name.split('.')
    if len(name_lst) > 1:
        new_obj = getattr(obj, name_lst[0])
        RecSetattr(new_obj, '.'.join(name_lst[1:]), value)
    else:
        setattr(obj, name, value)


"""
Jacob Thompson

MakeSet.py
"""

import numpy as np

def Line(m, b, num = 50, noise_deviation = 1, span = (-10, 10)):
	"""
	Makes a set of x's and y's in the shape of a parabola following the equation:
		y = m * x + b
	
	Adds noise and returns both x and y
	
	Note: The returned values are sorted in ascending order by the x's
	"""
	x = np.random.rand(num, 1) * (span[1] - span[0]) + span[0]
	x[:, 0].sort()
	noise = np.random.randn(num, 1) * noise_deviation
	y = x @ [[m]] + b + noise
	return x, y
	
def Parabola(a, b, c, num = 50, noise_deviation = 1, span = (-10, 10)):
	"""
	Makes a set of x's and y's shaped in a parabola of the form:
		y = ax^2 + bx + c

	Adds noise and returns both x and y
	"""
	x = np.random.rand(num, 1) * (span[1] - span[0]) + span[0]	# Generates random smattering of x's
	x[:, 0].sort()												# Sort the x's
	noise = np.random.randn(num, 1) * noise_deviation			# Generate a column of random noise
	X = np.hstack((np.ones((num, 1)), x, x ** 2))				# Generate the X array by combining ones, x, and x^2
	w = [[c], [b], [a]]											# Builds the weight vector from a, b, and c
	y = X @ w + noise											# Calculates y
	
	return x, y
	
def MultiLine(weights, num = 50, noise_deviation = 1, span = (-10, 10)):
	if type(weights) != np.array:
		weights = np.array(weights)
	
	dims = weights.shape[0] - 1
	# print(dims)
	x = np.random.rand(num, dims) * (span[1] - span[0]) + span[0]
	bias = np.ones([num, 1])
	x = np.hstack((bias, x))
	noise = np.random.randn(num, 1) * noise_deviation
	# print('x:', x.shape)
	# print('w:', weights.shape)
	y = x @ weights + noise	
	# print('y:', y.shape)
	return x, y


def TriCluster(num = 6000, scale = 1, skew = False):
	if skew:
		cov1 = np.random.randn(2, 2)
		cov2 = np.random.randn(2, 2)
		cov3 = np.random.randn(2, 2)
	else:
		cov1 = np.eye(2)
		cov2 = np.eye(2)
		cov3 = np.eye(2)

	x1 = np.random.randn(num // 3, 2) * scale @ cov1 + np.array([[0, -6]])
	x2 = np.random.randn(num // 3, 2) * scale @ cov2 + np.array([[8, 0]])
	x3 = np.random.randn(num // 3, 2) * scale @ cov3 + np.array([[0, 6]])

	cat_1 = np.hstack([
		x1,
		np.ones([x1.shape[0], 1]),
		np.zeros([x1.shape[0], 1]),
		np.zeros([x1.shape[0], 1]),
	])
	cat_2 = np.hstack([
		x2,
		np.zeros([x1.shape[0], 1]),
		np.ones([x1.shape[0], 1]),
		np.zeros([x1.shape[0], 1]),
	])
	cat_3 = np.hstack([
		x3,
		np.zeros([x1.shape[0], 1]),
		np.zeros([x1.shape[0], 1]),
		np.ones([x1.shape[0], 1]),
	])

	data = np.vstack([cat_1, cat_2, cat_3])

	np.random.shuffle(data)

	x = data[:, [0, 1]]
	y = data[:, 2:]

	return x, y

def SpanSpace(x, num = 100):
	"""
	:param x: Input data arranged in columns.
	:param num: number of points along each axis to create.
	:return: Simulated input data that spans the input data on each features.

	Works like an n-dimensional np.linspace() except the upper/lower bounds are
	determined by the max/min of the data along each of it's features.
	"""
	step = 1 / num
	n = x.shape[1]
	grid = np.mgrid[[slice(0, 1, step)] * n]
	return np.reshape(grid, (n, -1)).T * (x.max(axis = 0) - x.min(axis = 0)) + x.min(axis = 0)


"""
Jacob Thompson

Gradients.py
"""

import numpy as np

class Gradient:
    def __init__(self):
        self.change = 0

    def Evaluate(self, grad, learning_rate):
        return grad * learning_rate

class Momentum(Gradient):
    def __init__(self, velocity_rate = 0.9):
        self.velocity_rate = velocity_rate

        self.velocity = 0
        self.change = 0

    def Evaluate(self, grad, learning_rate):
        self.velocity = self.velocity_rate * self.velocity + learning_rate * grad
        return self.velocity

class Ada_Grad(Gradient):
    def __init__(self, adaptive_rate = 1, epsilon = 1e-8):
        self.adaptive_rate = adaptive_rate
        self.epsilon = epsilon

        self.ada_grad = 0
        self.change = 0

    def Evaluate(self, grad, learning_rate):
        self.ada_grad = self.adaptive_rate * self.ada_grad + grad ** 2
        learning_rate = learning_rate / np.sqrt(self.ada_grad + self.epsilon)
        return learning_rate * grad

class Momentum_Ada_Grad(Gradient):
    def __init__(self, velocity_rate = 0.9, adaptive_rate = 1, epsilon = 1e-8):
        self.velocity_rate = velocity_rate
        self.adaptive_rate = adaptive_rate
        self.epsilon = epsilon

        self.velocity = 0
        self.ada_grad = 0
        self.change = 0

    def Evaluate(self, grad, learning_rate):
        self.ada_grad = self.adaptive_rate * self.ada_grad + grad ** 2
        learning_rate = learning_rate / np.sqrt(self.ada_grad + self.epsilon)
        self.velocity = self.velocity_rate * self.velocity + learning_rate * grad
        return self.velocity

class RMS_Prop(Gradient):
    def __init__(self, adaptive_rate = 0.5, epsilon = 1e-8):
        self.adaptive_rate = adaptive_rate
        self.epsilon = epsilon

        self.grad = 0

        
        self.change = 0

    def Evaluate(self, grad, learning_rate):
        self.grad = self.adaptive_rate * self.grad + (1 - self.adaptive_rate) * grad ** 2
        learning_rate = learning_rate / np.sqrt(self.grad + self.epsilon)
        return learning_rate * grad

class Momentum_RMS_Prop(Gradient):
    def __init__(self, velocity_rate = 0.9, adaptive_rate = 0.5, epsilon = 1e-8):
        self.velocity_rate = velocity_rate
        self.adaptive_rate = adaptive_rate
        self.epsilon = epsilon

        self.velocity = 0
        self.grad = 0

        
        self.change = 0

    def Evaluate(self, grad, learning_rate):
        self.grad = self.adaptive_rate * self.grad + (1 - self.adaptive_rate) * grad ** 2
        learning_rate = learning_rate / np.sqrt(self.grad + self.epsilon)
        self.velocity = self.velocity_rate * self.velocity + learning_rate * grad
        return self.velocity

class Nesterov_Momentum(Gradient):
    def __init__(self, velocity_rate = 0.9):
        self.velocity_rate = velocity_rate
        self.velocity = 0

        
        self.change = 0

    def Evaluate(self, grad, learning_rate):
        out = -self.change
        self.change = self.velocity_rate * self.change + learning_rate * grad
        return out + self.change * 2


class Nesterov_RMS_Prop(Gradient):
    def __init__(self, velocity_rate = 0.9, adaptive_rate = 0.5, epsilon = 1e-8):
        self.velocity_rate = velocity_rate
        self.adaptive_rate = adaptive_rate
        self.epsilon = epsilon

        self.velocity = 0
        self.grad = 0
        
        self.change = 0

    def Evaluate(self, grad, learning_rate):
        out = -self.change

        self.grad = self.adaptive_rate * self.grad + (1 - self.adaptive_rate) * grad ** 2
        learning_rate = learning_rate / np.sqrt(self.grad + self.epsilon)

        self.change = self.velocity_rate * self.change + learning_rate * grad
        return out + self.change * 2

class Adam(Gradient):
    def __init__(self, velocity_rate = 0.9, adaptive_rate = 0.5, epsilon = 1e-8):
        self.velocity_rate = velocity_rate
        self.adaptive_rate = adaptive_rate
        self.epsilon = epsilon

        self.velocity = 0
        self.grad = 0

        self.i = 0

        self.change = 0

    def Evaluate(self, grad, learning_rate):
        self.i += 1

        self.grad = self.adaptive_rate * self.grad + (1 - self.adaptive_rate) * grad ** 2
        grad_hat = self.grad / (1 + self.adaptive_rate ** self.i)
        learning_rate = learning_rate / np.sqrt(grad_hat + self.epsilon)

        self.velocity = self.velocity_rate * self.velocity + (1 - self.velocity_rate) * grad
        velocity_hat = self.velocity / (1 + self.velocity_rate ** self.i)

        return learning_rate * velocity_hat


class Nesterov_Adam(Gradient):
    def __init__(self, velocity_rate = 0.9, adaptive_rate = 0.5, epsilon = 1e-8):
        self.velocity_rate = velocity_rate
        self.adaptive_rate = adaptive_rate
        self.epsilon = epsilon

        self.velocity = 0
        self.grad = 0
        self.out = 0

        self.i = 0

        self.change = 0

    def Evaluate(self, grad, learning_rate):
        self.i += 1

        out = -self.change

        self.grad = self.adaptive_rate * self.grad + (1 - self.adaptive_rate) * grad ** 2
        grad_hat = self.grad / (1 + self.adaptive_rate ** self.i)
        learning_rate = learning_rate / np.sqrt(grad_hat + self.epsilon)

        self.velocity = self.velocity_rate * self.velocity + (1 - self.velocity_rate) * grad
        velocity_hat = self.velocity / (1 + self.velocity_rate ** self.i)

        self.change = learning_rate * velocity_hat

        return out + self.change * 2


class Nesterov_Ada_Grad(Gradient):
    def __init__(self, adaptive_rate = 1, epsilon = 1e-8):
        self.adaptive_rate = adaptive_rate
        self.epsilon = epsilon

        self.ada_grad = 0

    def Evaluate(self, grad, learning_rate):
        out = -self.change

        self.ada_grad = self.adaptive_rate * self.ada_grad + grad ** 2
        learning_rate = learning_rate / np.sqrt(self.ada_grad + self.epsilon)

        self.change = self.velocity_rate * self.change + learning_rate * grad
        return out + self.change * 2




"""
Jacob Thompson
NNet.py
"""

import numpy as np
import matplotlib.pyplot as plot

import JT.Layers

class Neural_Network:
	def __init__(self, x, y, loss_func, echo = False):
		self.layer_list = []
		self.x = x
		self.y = y
		
		self.loss_func = loss_func
		self.activation = self.loss_func.activation
		
		self.input_layer = JT.Layers.Input_Layer(x.shape[1])
		self.output_layer = JT.Layers.Output_Layer(y.shape[1], self.activation)
	
		self.echo = echo
		
	def Echo(self, *args):
		if self.echo:
			args = list(args)
			for i in range(len(args)):
				if type(args[i]) != 'str':
					args[i] = str(args[i])
			print(' '.join(args))
	
	def AddLayer(self, layer):
		if len(self.layer_list) == 0:
			self.input_layer.next_layer = layer
			layer.prev_layer = self.input_layer
		else:
			self.layer_list[-1].next_layer = layer
			layer.prev_layer = self.layer_list[-1]
		
		layer.InitWeights()
		layer.next_layer = self.output_layer
		self.output_layer.prev_layer = layer
		self.output_layer.InitWeights()
		self.layer_list.append(layer)
		
	def Forward(self, sample_indices = None):
		if sample_indices is None:
			sample_x = self.x
		else:
			sample_x = self.x[sample_indices, :]
			
		self.p_hat = self.input_layer.Forward(sample_x)

	def Backward(self, learn_rate, sample_indices = None):
		if sample_indices is None:
			sample_y = self.y
		else:
			sample_y = self.y[sample_indices, :]
		self.output_layer.Backward(self.p_hat, sample_y, learn_rate)
		
	def Build(self, layer_list):
		for layer in layer_list:
			self.AddLayer(layer)
			
	def Fit(self, max_epochs, learn_rate, init_weights = True):
		if init_weights:
			self.loss = []
			for layer in self.layer_list:
				layer.InitWeights()
			self.output_layer.InitWeights()
			
		for i in range(max_epochs):
			self.Forward()
			self.Backward(learn_rate)
			p_hat = self.Predict(self.x)
			self.loss.append(self.loss_func(self.y, p_hat))
			self.Echo('Iteration:', i, '\tLoss:', self.loss[-1])
	
	def BatchFit(self, max_epochs, learn_rate, init_weights = True, batch_size = 10):
		if init_weights:
			self.loss = []
			for layer in self.layer_list:
				layer.InitWeights()
			self.output_layer.InitWeights()
			
		for i in range(max_epochs):
			sample_indices = (np.random.rand(batch_size) * self.x.shape[0]).astype(int)
			self.Forward(sample_indices = sample_indices)
			self.Backward(learn_rate, sample_indices = sample_indices)
			p_hat = self.Predict(self.x)
			self.loss.append(self.loss_func(self.y, p_hat))
			self.Echo('Iteration:', i, '\tLoss:', self.loss[-1])
	
	def Predict(self, x):
		p_hat = self.input_layer.Predict(x)
		return p_hat
			
	def PlotLoss(self):
		plot.plot(self.loss, label = 'Model Loss')
			
		plot.title('Loss Curve')
		plot.xlabel('Iterations')
		plot.ylabel('Loss')
		plot.legend()
		plot.show()
		



'''
Jacob Thompson

Feature Interaction
'''
import numpy as np
import pandas as pd

class Interaction():
	def Eval(self, feature):
		return self.Evaluate(feature)

class Inflation_Adjust(Interaction):
	def __init__(self, data_1, data_2):
		self.name = 'Adjust for Inflation'
		self.suffix = '_adjusted'
		cpi_data = pd.read_excel('CPI.xlsx', sheet_name = 'Data').to_numpy()
		self.to_current_dollars = cpi_data[:, [-1]]
		self.years = cpi_data[:, [1]]
		
	def Evaluate(self, value, year):
		return (year == self.years.T) @ self.to_current_dollars * value
		
class Multiply(Interaction):
	def __init__(self, data_1, data_2):
		self.name = 'Multiply'
		self.suffix = '_multiplied'
	
	def Evaluate(self, feature_1, feature_2):
		return (feature_1 * feature_2)


'''
Jacob Thompson

Feature Transforms
'''
import numpy as np

def StripedStack(array1, array2):
	if array1.shape != array2.shape:
		raise NameError('ARRAY SIZES DO NOT MATCH')
	out = np.array([])
	out.reshape(array1.shape[0], 0)
	for i in range(array1.shape[1]):
		out = np.hstack([out, array1[:, [i]], array2[:, [i]]])
	return out

class Transform():
	def Eval(self, feature):
		return self.Evaluate(feature)
	
	def Rev(self, feature):
		if 'Reverse' in dir(self):
			return self.Reverse(feature)
	
class Normalize(Transform):
	def __init__(self, data):
		self.name = 'Normalize'
		self.max = data.max()
		self.min = data.min()
		
	def Evaluate(self, feature):
		feature = (feature - self.min) / (self.max - self.min)
		return feature

	def Reverse(self, feature):
		feature * (self.max - self.min) + self.min
		return feature
		
class Square(Transform):
	def __init__(self, data):
		self.name = 'Square'
		self.suffix = '_squared'
		
	def Evaluate(self, feature):
		squared_feature = feature * feature
		return StripedStack(feature, feature_squared)
		
class Handle_Missing(Transform):
	def __init__(self, data, missing_str = 'NULL'):
		self.name = 'Handle Missing'
		self.suffix = '_missing'
		self.missing_str = missing_str
		
	def Evaluate(self, feature):
		feature_missing = feature == self.missing_str
		return StripedStack(feature, feature_missing)

class Category_Encode(Transform):
	def __init__(self, data):
		self.name = 'Category Encode'
		self.suffix = '_categorized'
		self.categories = list(set(data.flatten()))	
		
	def Evaluate(self, feature):
		feature_categorized = feature == categories
		return feature_categorized

class Binned_Encode(Transform):
	def __init__(self, data, bin_count = 50):
		bin_bounds = np.linspace(data.min(), data.max(), bin_count + 1)
		self.bin_bounds_lower = bin_bounds[:-1]
		self.bin_bounds_upper = bin_bounds[1:]
		self.bin_bounds_upper[-1] = np.inf
		
	def Evaluate(self, feature):
		feature_binned = (feature >= self.bin_bounds_lower) & (feature < self.bin_bounds_upper)
		return feature_binned
		

		
 
		
		
		
		
		


"""
Jacob Thompson

LossFunctions.py
"""
import numpy as np
import JT.ActivationFunctions

class Loss:
	"""
	Parent class that holds aliases
	"""
	def __init__(self):
		self.hyperparameters = {}
	
	def __init_subclass__(cls):
		cls.Eval = cls.Evaluate
		cls.__call__ = cls.Evaluate
		
		cls.D = cls.Derivative
		
	def SetHyperParameters(self, param_dict):
		next_dict = dict()
		for key, value in param_dict.items():
			if key in self.hyperparameters:
				setattr(self, key, value)
			else:
				next_dict[key] = value
		self.activation.SetHyperParameters(next_dict)
		
	def GetHyperParameters(self):
		out = {}
		for param in self.hyperparameters:
			out[param] = getattr(self, param)
		out.update(self.activation.GetHyperParameters())
		return out
			
class SumOfSquaredResidual(Loss):
	"""
	Class that contains the formula for calculating the L1 and L2 Penalized Sum of Squared Residual (aka Sum of Squared Error) (aaka ElasticNet)
	"""
	def __init__(self, reg_rate = 0, l1_ratio = 0):
		"""
		Initializes the object and stores the parameter for:
			lambda_1 - the penalty modifier for L1 penalization
			lambda_2 - the penalty modifier for L2 penalization
		"""
		self.reg_rate = reg_rate
		self.l1_ratio = l1_ratio
		self.lamb_1 = reg_rate * l1_ratio
		self.lamb_2 = reg_rate * (1-l1_ratio)
		self.hyperparameters = {'reg_rate', 'l1_ratio'}
		self.name = 'L1 and L2 Penalized Sum of Squared Residuals'
		self.activation = JT.ActivationFunctions.Activation
		
	def Evaluate(self, y, y_hat, w_hat = None):
		"""
		Evaluates the loss function given all the arguments
		"""
		
		if not w_hat is None:
			penalty = (
						self.lamb_1 * np.sum(abs(w_hat)) + 								# This is the formula for the L1 penalty
						self.lamb_2 * (w_hat.T @ w_hat).trace()							# This is the formula for the L2 penalty
						)
		else:
			penalty = 0
			
		return (
			((y - y_hat).T @ (y - y_hat)).trace() + 	# This is the formula for the typical Sum of Squared Residual
			penalty)
			
	def Derivative(self, w_hat, x, y, y_hat):
		"""
		Evaluates the derivative of the loss function given all the arguments
		"""
		return (
			x.T @ (y_hat - y) + 				# This is the formula for the typical Sum of Squared Residual derivative
			self.lamb_1 * np.sign(w_hat) + 		# This is the formula for the L1 penalty derivative
			self.lamb_2 * w_hat					# This is the formula for the L2 penalty derivative
			)
			
class SumOfSquaredError(SumOfSquaredResidual):
	"""
	Alias for L1L2SumOfSquaredResidual
	"""
	pass
			
class ElasticNet(SumOfSquaredResidual):
	"""
	Alias for L1L2SumOfSquaredResidual
	"""
	pass
	
class Binary_Cross_Entropy(Loss):
	def __init__(self, reg_rate = 0, l1_ratio = 0):
		self.name = 'Binary Cross Entropy'
		self.reg_rate = reg_rate
		self.l1_ratio = l1_ratio
		self.lamb_1 = reg_rate * l1_ratio
		self.lamb_2 = reg_rate * (1-l1_ratio)
		self.activation = JT.ActivationFunctions.Sigmoid()
		self.hyperparameters = {'reg_rate', 'l1_ratio'}
	
	def Evaluate(self, y, y_hat, w_hat = None):
		p_hat = self.activation(y_hat)
		
		if not (w_hat is None):
			penalty = (
						self.lamb_1 * np.sum(abs(w_hat)) + 								# This is the formula for the L1 penalty
						self.lamb_2 * (w_hat.T @ w_hat).trace()							# This is the formula for the L2 penalty
						)
		else:
			penalty = 0
			
		return (
			-np.sum(y * np.log(p_hat) + (1 - y) * np.log(1 - p_hat)) +
			self.lamb_1 * np.sum(abs(w_hat)) + 								# This is the formula for the L1 penalty
			self.lamb_2 * (w_hat.T @ w_hat).trace()	
			)
		
	def Derivative(self, w_hat, x, y, y_hat):
		p_hat = self.activation(y_hat)
		
		return (
			x.T @ (p_hat - y) + 
			self.lamb_1 * np.sign(w_hat) + 						# This is the formula for the L1 penalty derivative
			self.lamb_2 * w_hat	
			)
			
	# def Core(self, y_hat):
		# p_hat = self.activation(y_hat)
		# p_hat += (p_hat < 0.001) * 0.001
		# p_hat -= (p_hat > 0.999) * 0.001
		# return p_hat - y
	
class Cross_Entropy(Loss):
	def __init__(self, reg_rate = 0, l1_ratio = 0):
		self.name = 'Cross Entropy'
		self.lamb_1 = reg_rate * l1_ratio
		self.lamb_2 = reg_rate * (1-l1_ratio)
		self.activation = JT.ActivationFunctions.Softmax()
		self.hyperparameters = {'reg_rate', 'l1_ratio'}
		
	def Evaluate(self, y, y_hat, w_hat = None):
		p_hat = self.activation(y_hat)
		if not (w_hat is None):
			penalty = (
						self.lamb_1 * np.sum(abs(w_hat)) + 								# This is the formula for the L1 penalty
						self.lamb_2 * (w_hat.T @ w_hat).trace()							# This is the formula for the L2 penalty
						)
		else:
			penalty = 0
		return -np.sum(y * np.log(p_hat)) + penalty
			
			
	def Derivative(self, w_hat, x, y, y_hat):
		p_hat = self.activation(y_hat)
		return (
			x.T @ (p_hat - y) + 
			self.lamb_1 * np.sign(w_hat) + 						# This is the formula for the L1 penalty derivative
			self.lamb_2 * w_hat									# This is the formula for the L2 penalty derivative
			)
			


class Cross_Entropy2(Loss):
	def __init__(self, reg_rate = 0, l1_ratio = 0):
		self.name = 'Cross Entropy'
		self.lamb_1 = reg_rate * l1_ratio
		self.lamb_2 = reg_rate * (1-l1_ratio)
		self.activation = JT.ActivationFunctions.Softmax()
		self.hyperparameters = {'reg_rate', 'l1_ratio'}
		
	def Evaluate(self, y, p_hat, w_hat = None):
		if not (w_hat is None):
			penalty = (
						self.lamb_1 * np.sum(abs(w_hat)) + 								# This is the formula for the L1 penalty
						self.lamb_2 * (w_hat.T @ w_hat).trace()							# This is the formula for the L2 penalty
						)
		else:
			penalty = 0
		return -np.sum(y * np.log(p_hat)) + penalty
			
			
	def Derivative(self, w_hat, x, y, p_hat):
		return (
			x.T @ (p_hat - y) + 
			self.lamb_1 * np.sign(w_hat) + 						# This is the formula for the L1 penalty derivative
			self.lamb_2 * w_hat									# This is the formula for the L2 penalty derivative
			)



"""
Jacob Thompson

Activation Functions
"""
import numpy as np

class Activation:
	def __init__(self):
		self.hyperparameters = {}
		self.Eval = self.Evaluate
		self.__call__ = self.Evaluate

		self.D = self.Derivative
		
	def __init_subclass__(cls):
		cls.Eval = cls.Evaluate
		cls.__call__ = cls.Evaluate
		
		cls.D = cls.Derivative
		
	def SetHyperParameters(self, param_dict):
		next_dict = dict()
		for key, value in param_dict.items():
			if key in self.hyperparameters:
				setattr(self, key, value)
			else:
				next_dict[key] = value
				
	def GetHyperParameters(self):
		out = {}
		for param in self.hyperparameters:
			out[param] = getattr(self, param)
		return out
		
	def Evaluate(self, z):
		return z

	def Derivative(self, z):
		return 1
				
	
class Sigmoid(Activation):
	def Evaluate(self, z):
		"""
		Evaluates the sigmoid of z:
			1 / (1 + e^(-z))
		"""
		p_hat = 1 / (1 + np.exp(-z))
		p_hat += (p_hat < 0.0001) * 0.0001
		p_hat -= (p_hat > 0.9999) * 0.0001
		return p_hat
		
	def Derivative(self, z):
		"""
		Evaluates the derivative of sigmoid of z:
			Sigmoid(z) * (1 - Sigmoid(z))
		"""
		sig = self.Evaluate(z)
		return sig * (1 - sig)

class Softmax(Activation):
	def Evaluate(self, h):
		e_h = np.exp(h)
		p_hat = e_h / np.sum(e_h, axis = 1, keepdims = True)
		p_hat += (p_hat < 0.0001) * 0.0001
		p_hat -= (p_hat > 0.9999) * 0.0001
		return p_hat
		
	def Derivative(self, h):
		"""
		=== WIP ===
		"""
		pass
		
		
class Tanh(Activation):
	def Evaluate(self, h):
		return (np.exp(h) - np.exp(-h)) / (np.exp(h) + np.exp(-h))
		
	def Derivative(self, h):
		return 1 - (self.Evaluate(h) ** 2)
		
		
class ReLU(Activation):
	def Evaluate(self, h):
		return h * (h > 0)
		
	def Derivative(self, h):
		return h > 0
		
		
class LeakyReLU(Activation):
	def __init__(self, leak_coefficient):
		self.leak_coefficient = leak_coefficient
		self.hyperparameters = {'leak_coefficient'}
		
	def Evaluate(self, h):
		return (h * (h > 0)) + (self.leak_coefficient * h * (h < 0))
		
	def Derivative(self, h):
		return (h > 0) + (self.leak_coefficient * (h < 0))
		
class Softplus(Activation):
	def Evaluate(self, h):
		return np.log(1 + np.exp(h))
		
	def Derivative(self, h):
		return 1 / (1 + np.exp(-h))


"""
Jacob Thompson

Trees.py
"""


from copy import copy
import numpy as np
import scipy as sp
import scipy.stats

def Gini(y, axis = 0):
    out = 0
    for cat in np.unique(y):
        out += np.sum(y == cat, axis = axis, keepdims = True) / y.shape[0] + (1 - np.sum(y == cat, axis = axis, keepdims = True) / y.shape[0])
    return out


def InformationGain(x_k, y, t):
    z_0 = x_k < t
    z_1 = x_k >= t
    return Gini(y) - (Gini(y[np.where(z_0)]) * np.sum(z_0) + Gini(y[np.where(z_1)]) * np.sum(z_1)) / y.shape[0]

def Mode(z, axis = 0):
    return sp.stats.mode(z, axis = axis).mode


class Tree:
    def __init__(self, error_func=Gini, decision_func=Mode, stop_gain=0, max_depth=5, steps=-1):
        self.error_func = error_func
        self.decision_func = decision_func
        self.max_depth = max_depth
        self.stop_gain = stop_gain
        self.steps = steps

    def Train(self, x, y):
        self.x = x
        self.y = y
        self.size = y.shape[1]
        self.out = np.zeros(y.shape)
        if self.steps < 0:
            self.steps = self.x.shape[0]
        self.root = self.Node(self, np.where(np.ones(y.shape, dtype=bool))[0], self.max_depth)
        self.root.Split()
        del self.x
        del self.y
        return self.out

    def Predict(self, x):
        self.x = x
        self.out = np.zeros((x.shape[0], self.size))
        self.root.Predict(np.where(np.ones(x.shape[0], dtype=bool))[0])
        return self.out

    class Node:
        def __init__(self, parent, mask, height):
            self.size = np.sum(mask)
            self.height = height
            self.parent = parent
            if self.height == 0 or self.size == 1 or np.unique(self.y[mask]).size == 1:
                self.y_hat = self.decision_func(self.y[mask])
                self.out[mask] = self.y_hat
            else:
                self.mask = mask
                self.y_hat = None
                self.error = self.error_func(self.y[self.mask], axis = 0)

        def __getattr__(self, name):
            return self.parent.__getattribute__(name)

        def Split(self):
            if self.y_hat is None:
                max_gain = -np.inf
                y_mask = self.y[self.mask]
                opt_left = None
                opt_right = None
                for col in range(self.x.shape[1]):
                    x_col = self.x[self.mask, col]
                    for t in np.linspace(np.min(x_col), np.max(x_col), self.steps):
                        left = x_col > t
                        right = ~left

                        error_left = self.error_func(y_mask[left], axis = 0)
                        n_left = np.sum(left)

                        error_right = self.error_func(y_mask[right], axis = 0)
                        n_right = np.sum(right)

                        gain = self.error - (error_right * n_right + error_left * n_left) / self.size
                        #                         gains.append(gain)
                        if np.sum(gain) > max_gain:
                            opt_left = left
                            opt_right = right
                            self.opt_col = col
                            self.opt_t = t
                            max_gain = np.sum(gain)
                del x_col
                del y_mask
                if np.any(opt_left) and np.any(opt_right) and max_gain > self.stop_gain:
                    # print(self.mask)
                    # print(np.where(opt_right)[0])
                    mask_right = self.mask[np.where(opt_right)[0]]
                    mask_left = self.mask[opt_left]
                    del self.mask

                    self.left = self.Node(self.parent, mask_left, self.height - 1)
                    self.left.Split()
                    self.right = self.Node(self.parent, mask_right, self.height - 1)
                    self.right.Split()
                else:
                    self.y_hat = self.decision_func(self.y[self.mask], axis = 0)
                    self.out[self.mask] = self.y_hat
                    # print(self.y_hat)
                    del self.mask



        def Predict(self, mask):
            if self.y_hat is None:
                # print(mask)
                left = self.x[mask, self.opt_col] > self.opt_t
                right = ~left
                self.left.Predict(mask[np.where(left)[0]])
                self.right.Predict(mask[np.where(right)[0]])


            else:
                # print(self.y_hat)
                # print(mask.shape)
                self.out[mask] = self.y_hat


class Forest:
    def __init__(self, num, sample_size, subset_proportion=1, root_tree=Tree()):
        self.num = num
        self.root_tree = root_tree
        self.subset_proportion = subset_proportion
        self.sample_size = sample_size
        self.decision_func = self.root_tree.decision_func

    def Train(self, x, y):
        self.trees = []
        subsets = int(np.ceil(x.shape[1] * self.subset_proportion))
        self.subsets = []
        for i in range(self.num):
            self.subsets.append(np.random.choice(np.arange(x.shape[1]), size=subsets, replace=False))
            sample_indices = np.random.randint(0, x.shape[0], size=self.sample_size)
            self.trees.append(copy(self.root_tree))
            self.trees[-1].Train(x[sample_indices, :][:, self.subsets[-1]], y[sample_indices])

    def Predict(self, x):
        y_hats = []
        for i in range(self.num):
            y_hats.append(self.trees[i].Predict(x[:, self.subsets[i]]))

        y_hats = np.stack(y_hats)
        y_hat = self.decision_func(y_hats, axis=0)
        return y_hat


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="JT-Techfield",
    version="0.0.10",
    author="Jacob Thompson",
    author_email="jacob.thompson@techfield.us",
    description="My Data Science Module",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License", 
        "Operating System :: OS Independent",
    ],
	install_requires=[
		"numpy", 
		"matplotlib", 
		"pandas",
        "scipy",
        "torch"
		], 
	python_requires='>=3', 
	package_data = {
		'JT': ['*.xlsx'], 
	}, 
)


"""
Jacob Thompson

Weight Initialization
"""
import numpy as np


class Initializer:
	def __init__(self):
		self.Init = self.Initialize
		self.__call__ = self.Initialize

	def __init_subclass__(cls):
		cls.Init = cls.Initialize
		cls.__call__ = cls.Initialize
		
	def Initialize(self, size_in, size_out):
		return np.random.randn(size_in, size_out)


class Xavier(Initializer):
	def __init__(self, numerator = 1):
		"""
		numerator:	The value to use as the numerator for adjusting the weights
						Use 1 if using a ReLU activation
						Use 2 if using a tanh activation
		"""
		self.numerator = numerator
		
	def Initialize(self, size_in, size_out):
		return np.random.randn(size_in, size_out) * np.sqrt(self.numerator / size_in)


class He(Initializer):
	def Initialize(self, size_in, size_out):
		return np.random.randn(size_in, size_out) * np.sqrt(1 / (size_in + size_out))



"""
Jacob Thompson

Layers.py
"""

from collections import defaultdict
from copy import copy

import numpy as np
from scipy.signal import correlate

import JT.Initializers
import JT.Gradients

class Layer:
    def InitWeights(self):
        pass
    def Forward(self, z_in):
        pass
    def Backward(self, core_in, learn_rate):
        pass
    def Predict(self, z_in):
        pass

class Hidden_Layer(Layer):
    def __init__(self, activation, size, normalize_output = None, next_layer = None, prev_layer = None, batch_norm_parameter = 0.1, initializer = JT.Initializers.Initializer(), gradient = JT.Gradients.Gradient()):
        self.activation = activation
        self.next_layer = next_layer
        self.prev_layer = prev_layer
        self.size = size
        self.normalize_output = normalize_output
        self.InitWeights()
        self.mu_h = 0
        self.sig_h = 1
        self.alpha = batch_norm_parameter
        self.gradient = gradient
        self.initializer = initializer

        
    def Forward(self, z_in):
        self.z_in = z_in
        self.h = self.z_in @ self.weights + self.bias
        if self.normalize_output == 'before':
            mu_batch     = np.mean(self.h, axis = 0, keepdims = True)
            sig_batch    = np.std(self.h, axis = 0, keepdims = True)
            self.mu_h    = self.alpha * mu_batch  + (1 - self.alpha) * self.mu_h
            self.sig_h   = self.alpha * sig_batch + (1 - self.alpha) * self.sig_h
            self.denom   = np.sqrt(self.sig_h ** 2 + 1e-7)
            self.h_bar_0 = (self.h - self.mu_h) / self.denom
            self.h_bar_1 = self.gamma * self.h_bar_0 + self.beta
            self.z_out   = self.activation(self.h_bar_1)
            
        else:
            self.z_out = self.activation(self.h)
        
        if self.normalize_output == 'after':
            self.mu_z = np.mean(self.z_out, axis = 1, keepdims = True)
            self.sig_z = np.std(self.z_out, axis = 1, keepdims = True)
            self.denom = np.sqrt(self.sig_z ** 2 )
        
            self.z_out = (self.z_out - self.mu_z) / self.denom    
        return self.next_layer.Forward(self.z_out)
        
        
    def Predict(self, z_in):
        h = z_in @ (self.weights - self.gradients['weights'].change) + (self.bias - self.gradients['bias'].change)
        
        if self.normalize_output == 'before':
            h_bar_0 = (h - self.mu_h) / self.denom
            h_bar_1 = (self.gamma - self.gradients['gamma'].change) * h_bar_0 + (self.beta - self.gradients['beta'].change)
            z_out   = self.activation(h_bar_1)
        else:
            z_out = self.activation(h)
        
        if self.normalize_output == 'after':
            z_out = (z_out - self.mu_z) / self.denom
        
        return self.next_layer.Predict(z_out)
        
    def Backward(self, core_in, learn_rate):
        core = core_in * self.activation.D(self.h)
        
        
        if not (self.normalize_output is None):

            gamma_grad = np.sum(core * self.h_bar_0, axis = 0, keepdims = True)
            beta_grad = np.sum(core, axis = 0, keepdims = True)

            core *= (self.gamma / self.denom)
            
            self.gamma -= self.gradients['gamma'].Evaluate(gamma_grad, learn_rate)
            self.beta -= self.gradients['beta'].Evaluate(beta_grad, learn_rate)
            
            
        core_out = core @ self.weights.T
        
        weights_grad = self.z_in.T @ core
        bias_grad = np.sum(core, axis = 0, keepdims = True)
        self.weights -= self.gradients['weights'].Evaluate(weights_grad, learn_rate)
        self.bias -= self.gradients['bias'].Evaluate(bias_grad, learn_rate)
        
        self.prev_layer.Backward(core_out, learn_rate)
    
    def InitWeights(self):
        if not self.prev_layer is None:
        
            self.weights = self.initializer.Initialize(self.prev_layer.size, self.size)
            
            self.bias = np.random.randn(1, self.size)
            self.gamma = np.random.randn(1, self.size)
            self.beta = np.random.randn(1, self.size)

            self.gradients = defaultdict(lambda: copy(self.gradient))
        
    
        
class Output_Layer(Layer):
    def __init__(self, size, activation, prev_layer = None, initializer = JT.Initializers.Initializer()):
        self.size = size
        self.prev_layer = prev_layer
        self.activation = activation
        self.initializer = initializer
        
    def Forward(self, z_in):
        self.z_in = z_in
        self.h = self.z_in @ self.weights + self.bias
        self.p_hat = self.activation(self.h)
        return self.p_hat
        
    def Predict(self, z_in):
        h = z_in @ self.weights + self.bias
        p_hat = self.activation(h)
        return p_hat
        
    def Backward(self, p_hat, y, learn_rate):
        core = (p_hat - y)
        core_out = core @ self.weights.T
        self.weights -= learn_rate * self.z_in.T @ core
        self.bias -= learn_rate * np.sum(core, axis = 0, keepdims = True)
        
        self.prev_layer.Backward(core_out, learn_rate)
        
    def InitWeights(self):
        if not self.prev_layer is None:
            self.weights = self.initializer.Init(self.prev_layer.size, self.size)
            self.bias = np.random.randn(1, self.size)
        
class Input_Layer(Layer):
    def __init__(self, size, next_layer = None):
        self.size = size
        self.next_layer = next_layer
        
    def Forward(self, x):
        if x.shape[1] != self.size:
            raise NameError('Input not correct shape')
            
        return self.next_layer.Forward(x)
        
    def Backward(self, *args):
        pass
        
    def Predict(self, x):
        if x.shape[1] != self.size:
            raise NameError('Input not correct shape')
        
        return self.next_layer.Predict(x)

class Normalize_Layer(Layer):
    """
    ADD GAMMA AND BETA
    """
    def __init__(self, eta = 1e-8, next_layer = None, prev_layer = None):
        self.eta = eta
        self.next_layer = next_layer
        self.prev_layer = prev_layer
    
    def Forward(self, z_in):
        self.z_in = z_in
        self.mu = np.mean(self.z_in, axis = 1, keepdims = True)
        self.sig = np.std(self.z_in, axis = 1, keepdims = True)
        self.denominator = np.sqrt(self.sig ** 2 + self.eta)
        
        self.z_out = (self.z_in - self.mu) / self.denominator
        return self.next_layer.Forward(self.z_out)
        
    def Backward(self, core_in, learn_rate):
        core_out = core_in / self.denominator
        self.prev_layer.Backward(core_out, learn_rate)
        
    def Predict(self, z_in):
        self.z_out = (z_in - self.mu) / self.denominator
        return self.next_layer.Forward(self.z_out)
        
    def InitWeights(self):
        self.size = self.prev_layer.size
            
class Noise_Injector(Layer):
    def __init__(self, noise_param = 0.1, moving_avg_param = 0.1, next_layer = None, prev_layer = None):
        self.noise_param = noise_param
        self.moving_avg_param = moving_avg_param
        self.next_layer = next_layer
        self.prev_layer = prev_layer
        self.mu = 0
        self.sigma = 1
        
        
    def Forward(self, z_in):
        mu_batch     = np.mean(z_in, axis = 0, keepdims = True)
        sig_batch    = np.std(z_in, axis = 0, keepdims = True)
        
        self.mu      = self.moving_avg_param * mu_batch  + (1 - self.moving_avg_param) * self.mu
        self.sigma     = self.moving_avg_param * sig_batch + (1 - self.moving_avg_param) * self.sigma
        
        noise = self.noise_param * (np.random.randn(*z_in.shape) * self.sigma + self.mu)
        z_out = z_in + noise
        return self.next_layer.Forward(z_out)
        
    def Backward(self, core_in, learn_rate):
        self.prev_layer.Backward(core_in, learn_rate)
        
    def Predict(self, z_in):
        return self.next_layer.Predict(z_in)
        
    def InitWeights(self):
        self.size = self.prev_layer.size
        
class Dropout_Layer(Layer):
    def __init__(self, dropout_rate = 0.1, next_layer = None, prev_layer = None):
        self.dropout_rate = dropout_rate
        self.next_layer = next_layer
        self.prev_layer = prev_layer
        
    def Forward(self, z_in):
        z_out = z_in * (np.random.random(z_in.shape) > self.dropout_rate)
        return self.next_layer.Forward(z_out)
        
    def Backward(self, core_in, learn_rate):
        self.prev_layer.Backward(core_in, learn_rate)
        
    def Predict(self, z_in):
        z_out = z_in * (1 - self.dropout_rate)
        return self.next_layer.Predict(z_out)
        
    def InitWeights(self):
        self.size = self.prev_layer.size
        
class Convolution_Layer(Layer):
    def __init__(self, kernel_size, kernel_num, next_layer = None, prev_layer = None):
        self.kernel_size = kernel_size
        self.kernel_num = kernel_num
        self.next_layer = next_layer
        self.prev_layer = prev_layer

    def InitWeights(self):
        self.depth = self.prev_layer.kernel_num
        self.kernels = np.random.rand(self.kernel_num, self.kernel_size, self.kernel_size, self.depth)

    def Forward(self, z_in):
        out = []
        for kernel in self.kernels:
            out.append(correlate(z_in, kernel, mode = 'same')[:, :, (self.depth // 2):-(self.depth // 2)])
        return np.dstack(out)

    def Backward(self, core_in, learn_rate):
        # TODO: Write this backward method
        pass
    







